/// <reference types="node" />
import http, { AgentOptions, Agent, IncomingMessage, ClientRequest, ClientRequestArgs } from 'http';
import { Socket } from 'net';
import { EventEmitter } from 'stream';
import { URL } from 'url';

/**
 * Extended options for a fake agent.
 *
 * @public
 */
interface FakeAgentOptions extends AgentOptions {
    protocol?: string | null;
}
/**
 * Mock agent class to avoid request been sent.
 *
 * @public
 */
declare class FakeAgent extends EventEmitter implements Agent {
    options: AgentOptions;
    protocol: string | undefined | null;
    sockets: NodeJS.ReadOnlyDict<Socket[]>;
    freeSockets: NodeJS.ReadOnlyDict<Socket[]>;
    requests: NodeJS.ReadOnlyDict<IncomingMessage[]>;
    maxSockets: number;
    maxFreeSockets: number;
    maxTotalSockets: number;
    static is(value: unknown): value is FakeAgent;
    constructor(options?: FakeAgentOptions);
    destroy(): void;
    addRequest(): void;
}

/**
 * The options to mock when a retrieving fake request.
 *
 * @public
 */
interface RequestStubReturnOptions {
    status?: number;
    contentType?: string;
    response?: string;
    responseText?: string;
    responseHeaders?: {
        [key: string]: string;
    };
}

/**
 * The data to be used as the mock response of a fake request.
 *
 * @public
 */
interface Response {
    status?: number;
    statusText?: string;
    responseText?: string;
    response?: string;
    contentType?: string;
    responseHeaders?: {
        [key: string]: string;
    };
}

/**
 * Class that represents a mock response when a certain
 * request is made.
 *
 * @public
 */
declare class RequestStub implements Response {
    data: string | RegExp | null | undefined;
    method: string | null | undefined;
    status?: number;
    response?: string;
    contentType?: string;
    responseText?: string;
    responseHeaders?: {
        [p: string]: string;
    };
    private url;
    private query;
    constructor(url: string | RegExp, stubData?: string | RegExp | null, method?: string | null);
    andReturn(options: RequestStubReturnOptions): void;
    matches(fullUrl: string | RegExp, data?: string, method?: string): boolean;
    private _normalizeQuery;
}

/**
 * Tracker of all the requests made in the installed session.
 *
 * @public
 */
declare class RequestTracker {
    private requests;
    constructor();
    track(request: FakeRequest): void;
    first(): FakeRequest | undefined;
    mostRecent(): FakeRequest | undefined;
    at(index: number): FakeRequest | undefined;
    filter(urlToMatch: string | RegExp | ((request: FakeRequest) => boolean)): FakeRequest[];
    count(): number;
    reset(): void;
}

/**
 * Tracker of all the mock responses set.
 *
 * @public
 */
declare class StubTracker {
    private stubs;
    constructor();
    addStub(stub: RequestStub): void;
    findStub(url: string, data?: string, method?: string): RequestStub | undefined;
    reset(): void;
}

/**
 * Main class the manages the fake requests and the mock
 * responses that can be retrieved in a matching request.
 *
 * @public
 */
declare class MockAjax {
    readonly stubs: StubTracker;
    readonly requests: RequestTracker;
    private modules;
    private requestOverrides?;
    constructor(modules: {
        [protocol: string]: {
            request: typeof http.request;
        };
    });
    install(): void;
    uninstall(): void;
    stubRequest(url: string | RegExp, data?: string | RegExp | null, method?: string | null): RequestStub;
    private _mockRequest;
}

/**
 * Fake {@link ClientRequest} to be used instead a request call.
 *
 * @public
 */
declare class FakeRequest extends ClientRequest {
    url: string;
    socket: Socket;
    method: string;
    params: string | undefined;
    requestHeaders: {
        [key: string]: string;
    };
    private ajax;
    private res;
    private callback;
    private ended;
    private requestBodyBuffers;
    private responseBodyBuffers;
    constructor(ajax: MockAjax, optionsOrURL: ClientRequestArgs | string | URL, callback?: (res: IncomingMessage) => void);
    private static createURL;
    respondWith(response: Response): void;
    getHeaders(): FakeRequest['requestHeaders'];
    flushHeaders(): void;
    abort(): void;
    end(chunk?: unknown, encoding?: unknown): this;
    write(chunk: unknown, encoding?: unknown): boolean;
    once(event: string, listener: (...args: any[]) => void): this;
    on(event: string, listener: (...args: any[]) => void): this;
    private _processSocketListener;
    private _responseWithStub;
    private _endFake;
    private _emitError;
}

declare global {
    namespace jasmine {
        const Ajax: MockAjax;
    }
}

export { FakeAgent, type FakeAgentOptions, FakeRequest, MockAjax, RequestStub, type RequestStubReturnOptions, RequestTracker, type Response, StubTracker };
