'use strict';

var tslib = require('tslib');
var stream = require('stream');
var http = require('http');
var net = require('net');
var timers = require('timers');
var url = require('url');
var https = require('https');

/**
 * Mock agent class to avoid request been sent.
 *
 * @public
 */
var FakeAgent = /** @class */ (function (_super) {
    tslib.__extends(FakeAgent, _super);
    function FakeAgent(options) {
        if (options === void 0) { options = {}; }
        var _this = _super.call(this) || this;
        _this.protocol = options.protocol;
        _this.options = options;
        _this.sockets = {};
        _this.freeSockets = {};
        _this.requests = {};
        _this.maxSockets = Infinity;
        _this.maxFreeSockets = Infinity;
        _this.maxTotalSockets = Infinity;
        return _this;
    }
    FakeAgent.is = function (value) {
        return typeof value === 'object'
            && value !== null
            && 'addRequest' in value;
    };
    FakeAgent.prototype.destroy = function () {
        this.sockets = {};
        this.requests = {};
    };
    // Methods Node.js require to be recognized as Agent like.
    // eslint-disable-next-line class-methods-use-this
    FakeAgent.prototype.addRequest = function () {
        // To nothing with the request
    };
    return FakeAgent;
}(stream.EventEmitter));

/**
 * Fake {@link ClientRequest} to be used instead a request call.
 *
 * @public
 */
var FakeRequest = /** @class */ (function (_super) {
    tslib.__extends(FakeRequest, _super);
    function FakeRequest(ajax, optionsOrURL, callback) {
        var _this = this;
        var _a;
        var options = typeof optionsOrURL === 'object'
            ? optionsOrURL
            : new url.URL(optionsOrURL, 'http://localhost');
        if (!options.protocol && FakeAgent.is(options.agent)) {
            options.protocol = options.agent.protocol;
        }
        options.agent = new FakeAgent({ protocol: options.protocol });
        _this = _super.call(this, options) || this;
        _this.ajax = ajax;
        _this.callback = callback;
        _this.aborted = false;
        _this.socket = new net.Socket();
        _this.res = new http.IncomingMessage(_this.socket);
        _this.requestBodyBuffers = [];
        _this.responseBodyBuffers = [];
        _this.url = FakeRequest.createURL(options);
        // this.method = 'method' in options ? options.method ?? 'GET' : 'GET',
        _this.requestHeaders = _this.getHeaders();
        if (!((_a = options.headers) === null || _a === void 0 ? void 0 : _a.host)) {
            delete _this.requestHeaders.host;
        }
        return _this;
    }
    FakeRequest.createURL = function (options) {
        var url = '';
        if (options.protocol) {
            url += "".concat(options.protocol, "//");
        }
        if (options.host) {
            url += options.host;
        }
        else if (options.hostname) {
            url += options.hostname;
            if (options.port) {
                url += ':' + options.port;
            }
        }
        if ('path' in options) {
            url += options.path;
        }
        else if ('pathname' in options) {
            url += options.pathname;
        }
        url = url.replace(/\+/g, ' ');
        return decodeURIComponent(url);
    };
    FakeRequest.prototype.respondWith = function (response) {
        var _a, _b;
        if (this.ended)
            return;
        this.res.statusCode = (_a = response.status) !== null && _a !== void 0 ? _a : 200;
        this.res.headers = response.responseHeaders ? tslib.__assign({}, response.responseHeaders) : {};
        var responseData = (_b = response.response) !== null && _b !== void 0 ? _b : response.responseText;
        if (responseData) {
            this.responseBodyBuffers.push(responseData);
        }
        this._endFake();
    };
    // Overridden methods
    FakeRequest.prototype.getHeaders = function () {
        var headers = {};
        for (var _i = 0, _a = this.getHeaderNames(); _i < _a.length; _i++) {
            var name_1 = _a[_i];
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            var header = this.getHeader(name_1);
            headers[name_1] = header.toString();
        }
        return headers;
    };
    FakeRequest.prototype.flushHeaders = function () {
        this.end();
    };
    FakeRequest.prototype.abort = function () {
        if (this.aborted)
            return;
        this.aborted = true;
        if (!this.ended)
            this._endFake();
        var error = new Error();
        error.code = 'aborted';
        this.res.emit('close', error);
        this.socket.destroy();
        this.emit('abort');
        var connResetError = new Error('socket hang up');
        connResetError.code = 'ECONNRESET';
        this._emitError(connResetError);
    };
    FakeRequest.prototype.end = function (chunk, encoding) {
        if (this.aborted)
            this._emitError(new Error('Request aborted'));
        if (this.ended)
            return this;
        if (chunk)
            this.write(chunk, encoding);
        this._endFake();
        this.emit('finish');
        this.emit('end');
        return this;
    };
    FakeRequest.prototype.write = function (chunk, encoding) {
        var _this = this;
        if (encoding === void 0) { encoding = 'utf8'; }
        if (typeof chunk !== 'string' && !Buffer.isBuffer(chunk))
            throw new TypeError('Unsupported chunk type');
        if (chunk && !this.aborted) {
            if (!Buffer.isBuffer(chunk)) {
                if (typeof encoding !== 'string')
                    throw new TypeError('Unsupported encode type');
                chunk = Buffer.from(chunk, encoding);
            }
            this.requestBodyBuffers.push(chunk);
        }
        if (this.aborted)
            this._emitError(new Error('Request aborted'));
        timers.setImmediate(function () { return _this.emit('drain'); });
        return false;
    };
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    FakeRequest.prototype.once = function (event, listener) {
        if (event === 'socket') {
            return this._processSocketListener(listener);
        }
        return _super.prototype.once.call(this, event, listener);
    };
    // eslint-disable-next-line @typescript-eslint/no-explicit-any
    FakeRequest.prototype.on = function (event, listener) {
        if (event === 'socket') {
            return this._processSocketListener(listener);
        }
        return _super.prototype.on.call(this, event, listener);
    };
    FakeRequest.prototype._processSocketListener = function (listener) {
        listener.call(this, this.socket);
        this.socket.emit('connect', this.socket);
        this.socket.emit('secureConnect', this.socket);
        return this;
    };
    FakeRequest.prototype._responseWithStub = function () {
        if (this.ended)
            return;
        var stub = this.ajax.stubs.findStub(this.url, this.params, this.method);
        if (stub)
            this.respondWith(stub);
        this.ended = true;
    };
    FakeRequest.prototype._endFake = function () {
        var _this = this;
        if (this.params === void 0 && this.requestBodyBuffers.length) {
            var requestBodyBuffer = Buffer.concat(this.requestBodyBuffers);
            //  When request body is a binary buffer we internally use in its hexadecimal representation.
            this.params = isBinaryBuffer(requestBodyBuffer)
                ? requestBodyBuffer.toString('hex')
                : requestBodyBuffer.toString('utf8');
        }
        // Try to response with stub if no previous response
        if (this.res.statusCode === null)
            return this._responseWithStub();
        this.ended = true;
        if (this.aborted)
            return;
        process.nextTick(function () {
            if (_this.aborted)
                return;
            if (typeof _this.callback === 'function') {
                _this.callback.call(void 0, _this.res);
            }
            _this.emit('response', _this.res);
            // Stream the response chunks one at a time.
            var responseBodyBuffers = _this.responseBodyBuffers.concat();
            var emitChunk = function () {
                var chunk = responseBodyBuffers.shift();
                if (chunk === undefined) {
                    _this.res.push(null);
                    return;
                }
                _this.res.push(chunk);
                timers.setImmediate(emitChunk);
            };
            timers.setImmediate(emitChunk);
        });
    };
    FakeRequest.prototype._emitError = function (error) {
        var _this = this;
        process.nextTick(function () { return _this.emit('error', error); });
    };
    return FakeRequest;
}(http.ClientRequest));
function isBinaryBuffer(buffer) {
    //  Test if the buffer can be reconstructed verbatim from its utf8 encoding.
    var utfEncodedBuffer = buffer.toString('utf8');
    var reconstructedBuffer = Buffer.from(utfEncodedBuffer, 'utf8');
    //  If the buffers are *not* equal then this is a 'binary buffer'
    //  meaning that it cannot be faithfully represented in utf8.
    return !buffer.equals(reconstructedBuffer);
}

/**
 * Class that represents a mock response when a certain
 * request is made.
 *
 * @public
 */
var RequestStub = /** @class */ (function () {
    function RequestStub(url, stubData, method) {
        var _a;
        if (url instanceof RegExp) {
            this.url = url;
            this.query = void 0;
        }
        else {
            _a = url.split("?"), this.url = _a[0], this.query = _a[1];
            this.query = this._normalizeQuery(this.query);
        }
        this.data = stubData;
        this.method = method;
    }
    RequestStub.prototype.andReturn = function (options) {
        this.status = options.status || 200;
        this.contentType = options.contentType;
        this.response = options.response;
        this.responseText = options.responseText;
        this.responseHeaders = options.responseHeaders;
    };
    RequestStub.prototype.matches = function (fullUrl, data, method) {
        fullUrl = fullUrl.toString();
        var urlMatches = false;
        if (this.url instanceof RegExp) {
            urlMatches = this.url.test(fullUrl);
        }
        else {
            var _a = fullUrl.split("?"), url = _a[0], query = _a[1];
            urlMatches = this.url === url && this.query === this._normalizeQuery(query);
        }
        var dataMatches = false;
        if (this.data instanceof RegExp) {
            if (data)
                dataMatches = this.data.test(data);
        }
        else {
            dataMatches = !this.data || this.data === this._normalizeQuery(data);
        }
        var methodMatches = !this.method || this.method === method;
        return urlMatches && dataMatches && methodMatches;
    };
    RequestStub.prototype._normalizeQuery = function (query) {
        if (!query)
            return;
        return query
            .split("&")
            .sort()
            .join("&");
    };
    return RequestStub;
}());

/**
 * Tracker of all the requests made in the installed session.
 *
 * @public
 */
var RequestTracker = /** @class */ (function () {
    function RequestTracker() {
        this.requests = [];
    }
    RequestTracker.prototype.track = function (request) {
        this.requests.push(request);
    };
    RequestTracker.prototype.first = function () {
        return this.at(0);
    };
    RequestTracker.prototype.mostRecent = function () {
        return this.at(this.requests.length - 1);
    };
    RequestTracker.prototype.at = function (index) {
        if (index >= this.requests.length)
            return;
        return this.requests[index];
    };
    RequestTracker.prototype.filter = function (urlToMatch) {
        if (urlToMatch instanceof RegExp) {
            var urlRegex_1 = urlToMatch;
            urlToMatch = function (request) {
                return urlRegex_1.test(request.url);
            };
        }
        if (typeof urlToMatch === 'string') {
            var urlString_1 = new url.URL(urlToMatch).toString();
            urlToMatch = function (request) {
                return urlString_1 === request.url;
            };
        }
        return this.requests.filter(urlToMatch);
    };
    RequestTracker.prototype.count = function () {
        return this.requests.length;
    };
    RequestTracker.prototype.reset = function () {
        this.requests.length = 0;
    };
    return RequestTracker;
}());

/**
 * Tracker of all the mock responses set.
 *
 * @public
 */
var StubTracker = /** @class */ (function () {
    function StubTracker() {
        this.stubs = [];
    }
    StubTracker.prototype.addStub = function (stub) {
        this.stubs.unshift(stub);
    };
    StubTracker.prototype.findStub = function (url, data, method) {
        return this
            .stubs
            .find(function (stub) { return stub.matches(url, data, method); });
    };
    StubTracker.prototype.reset = function () {
        this.stubs.length = 0;
    };
    return StubTracker;
}());

/**
 * Main class the manages the fake requests and the mock
 * responses that can be retrieved in a matching request.
 *
 * @public
 */
var MockAjax = /** @class */ (function () {
    function MockAjax(modules) {
        this.modules = modules;
        this.stubs = new StubTracker();
        this.requests = new RequestTracker();
    }
    MockAjax.prototype.install = function () {
        if (this.requestOverrides)
            throw new Error('Mock Jasmine Ajax is already installed.');
        this.requestOverrides = {};
        for (var _i = 0, _a = Object.entries(this.modules); _i < _a.length; _i++) {
            var _b = _a[_i], protocol = _b[0], module_1 = _b[1];
            // Store original method
            this.requestOverrides[protocol] = module_1.request;
            // Replace request method
            module_1.request = this._mockRequest.bind(this);
        }
    };
    MockAjax.prototype.uninstall = function () {
        if (!this.requestOverrides)
            throw new Error('Mock Jasmine Ajax is not installed.');
        //  Restore any overridden requests.
        for (var _i = 0, _a = Object.entries(this.modules); _i < _a.length; _i++) {
            var _b = _a[_i], protocol = _b[0], module_2 = _b[1];
            module_2.request = this.requestOverrides[protocol];
        }
        this.requestOverrides = undefined;
        this.stubs.reset();
        this.requests.reset();
    };
    MockAjax.prototype.stubRequest = function (url, data, method) {
        var stub = new RequestStub(url, data, method);
        this.stubs.addStub(stub);
        return stub;
    };
    MockAjax.prototype._mockRequest = function (optionsOrUrl, callbackOrOptions, callback) {
        if (typeof callbackOrOptions === 'function')
            callback = callbackOrOptions;
        if (typeof callbackOrOptions === 'object') {
            if (typeof optionsOrUrl === 'string')
                optionsOrUrl = url.parse(optionsOrUrl);
            optionsOrUrl = tslib.__assign(tslib.__assign({}, optionsOrUrl), callbackOrOptions);
        }
        var request = new FakeRequest(this, optionsOrUrl, callback);
        this.requests.track(request);
        return request;
    };
    return MockAjax;
}());

var jasmineObj = jasmine;
jasmineObj.Ajax = new MockAjax({ http: http, https: https });

exports.FakeAgent = FakeAgent;
exports.FakeRequest = FakeRequest;
exports.MockAjax = MockAjax;
exports.RequestStub = RequestStub;
exports.RequestTracker = RequestTracker;
exports.StubTracker = StubTracker;
//# sourceMappingURL=index.js.map
